/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogEnvironment.h"
#include "DebuggerCoreInterface.h"
#include "Debugger.h"
#include "SymbolManager.h"
#include "Configuration.h"
#include <QStringListModel>
#include <QSortFilterProxyModel>

#include <QHeaderView>
#include <QMessageBox>
#include <QVector>
#include <QFileInfo>

//------------------------------------------------------------------------------
// Name: DialogEnvironment(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
DialogEnvironment::DialogEnvironment(QWidget *parent) : QDialog(parent) {
	ui.setupUi(this);
	
	m_Model = new QStringListModel(this);
	m_FilterModel = new QSortFilterProxyModel(this);
	
	m_FilterModel->setFilterKeyColumn(0);
	m_FilterModel->setSourceModel(m_Model);
	ui.listView->setModel(m_FilterModel);
	
	connect(ui.txtSearch, SIGNAL(textChanged(const QString &) ), m_FilterModel, SLOT(setFilterFixedString(const QString &)));
}

//------------------------------------------------------------------------------
// Name: on_listView_doubleClicked(const QModelIndex &index) 
// Desc: follows the found item in the data view
//------------------------------------------------------------------------------
void DialogEnvironment::on_listView_doubleClicked(const QModelIndex &index) {

	bool ok;
	const QString s = index.data().toString();
	
#if defined(EDB_X86)
		const edb::address_t addr = s.left(8).toULong(&ok, 16);
#elif defined(EDB_X86_64)
		const edb::address_t addr = s.left(16).toULong(&ok, 16);
#endif
	if(ok) {
		edb::v1::dumpData(addr);
	}
}

//------------------------------------------------------------------------------
// Name: getLibcName()
// Desc: 
//------------------------------------------------------------------------------
QString DialogEnvironment::getLibcName() const {

	QString libcName;

	const QStringList libs = edb::v1::loadedLibraries();

	foreach(QString s, libs) {
		if(!libcName.isEmpty()) {
			break;
		}

		const QFileInfo fileinfo(s);

		// this tries its best to cover all possible libc library versioning possibilities
		// we need to find out if this is 100% accurate, so far seems correct based on my system

		if(fileinfo.completeBaseName().startsWith("libc-")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			edb::v1::log("[Environment] libc library appears to be: %s", qPrintable(libcName));
			continue;
		}

		if(fileinfo.completeBaseName().startsWith("libc.so")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			edb::v1::log("[Environment] libc library appears to be: %s", qPrintable(libcName));
			continue;
		}
	}
	
	return libcName;
}

//------------------------------------------------------------------------------
// Name: doFind()
// Desc: 
//------------------------------------------------------------------------------
void DialogEnvironment::doFind() {
	QStringList results;

	const QString libcName = getLibcName();
	const SymbolManager::Symbol *const s = edb::v1::symbolManager().findSymbol(libcName + "::__environ");
	if(s != 0) {
		edb::address_t evironment_address = s->address;

		edb::v1::log("[Environment] pointer to environment array is at: " EDB_FMT_PTR, evironment_address);
		edb::v1::debuggerBase->readBytes(evironment_address, &evironment_address, sizeof(evironment_address));
		edb::v1::log("[Environment] environment array is at: " EDB_FMT_PTR, evironment_address);

		edb::address_t address;
		edb::v1::debuggerBase->readBytes(evironment_address, &address, sizeof(address));
		QString temp;
		while(address != 0) {

			QString str;
			int stringLen;
			const bool ok = edb::v1::getASCIIStringAtAddress(address, str, edb::v1::config().minAsciiStringLength, INT_MAX, stringLen);
			if(ok) {
				results << temp.sprintf(EDB_FMT_PTR ": %s", address, qPrintable(str));
			}

			evironment_address += sizeof(edb::address_t);
			edb::v1::debuggerBase->readBytes(evironment_address, &address, sizeof(address));
		}
	} else {
		//QMessageBox::information(this, "__environ symbol not found in libc", "Could not find the symbol for <strong>__environ</strong> in your libc, perhaps you need to regenerate your symbols?");
		edb::v1::log("[Environment] __environ symbol not found in libc");
	}
	
	m_Model->setStringList(results);
}

//------------------------------------------------------------------------------
// Name: on_btnRefresh_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogEnvironment::on_btnRefresh_clicked() {
	ui.btnRefresh->setEnabled(false);
	doFind();
	ui.btnRefresh->setEnabled(true);
}


//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: 
//------------------------------------------------------------------------------
void DialogEnvironment::showEvent(QShowEvent *) {
	on_btnRefresh_clicked();
}
