/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "ArchProcessor.h"
#include "QCatagoryList.h"
#include "Debugger.h"
#include "DebuggerCoreInterface.h"
#include "Instruction.h"
#include "State.h"
#include "Configuration.h"
#include "Util.h"

const ArchProcessor::Function ArchProcessor::funcs[] = {
	// initialize our system with some libc definitions
	{ "memchr", FunctionInfo('p', 'c', 'u') },
	{ "memcmp", FunctionInfo('p', 'p', 'u') },
	{ "memcpy", FunctionInfo('p', 'p', 'u') },
	{ "memmove", FunctionInfo('p', 'p', 'u') },
	{ "memset", FunctionInfo('p', 'c', 'u') },
	{ "strcat", FunctionInfo('s', 's') },
	{ "strchr", FunctionInfo('s', 'c') },
	{ "strcmp", FunctionInfo('s', 's') },
	{ "strcoll", FunctionInfo('s', 's') },
	{ "strcpy", FunctionInfo('p', 's') },
	{ "strcspn", FunctionInfo('s', 's') },
	{ "strerror", FunctionInfo('i') },
	{ "strlen", FunctionInfo('s') },
	{ "strncat", FunctionInfo('s', 's', 'u') },
	{ "strncmp", FunctionInfo('s', 's', 'u') },
	{ "strncpy", FunctionInfo('p', 's', 'u') },
	{ "strpbrk", FunctionInfo('s', 's') },
	{ "strrchr", FunctionInfo('s', 'c') },
	{ "strspn", FunctionInfo('s', 's') },
	{ "strstr", FunctionInfo('s', 's') },
	{ "strtok", FunctionInfo('s', 's') },
	{ "strxfrm", FunctionInfo('p', 's', 'u') },
	{ "strdup", FunctionInfo('s') },
	{ "strndup", FunctionInfo('s', 'u') },

	{ "fopen", FunctionInfo('s', 'p') },
	{ "fclose", FunctionInfo('p') },
	{ "fread", FunctionInfo('p', 'u', 'u', 'p') },
	{ "fwrite", FunctionInfo('p', 'u', 'u', 'p') },

	{ "free", FunctionInfo('p') },
	{ "exit", FunctionInfo('i') },
	{ "printf", FunctionInfo('f') },
	{ "fputc", FunctionInfo('c', 'p') },
	{ "fputs", FunctionInfo('s', 'p') },
	{ "putchar", FunctionInfo('c') },
	{ "puts", FunctionInfo('s') },
	{ "remove", FunctionInfo('s') },
	{ "setlocale", FunctionInfo('i', 's') },

	// other functions
	{ "sbrk", FunctionInfo('p') },
	{ "__errno_location", FunctionInfo() },
	{ "__libc_start_main", FunctionInfo('p', 'i', 'p', 'p', 'p', 'p', 'p') },
	{ "getpagesize", FunctionInfo() },
	{ "ioctl", FunctionInfo('i', 'i', 'p') },
	{ "isatty", FunctionInfo('i') },
	{ "poll", FunctionInfo('p', 'u', 'i') },
	{ "getenv", FunctionInfo('s') },
	{ "setenv", FunctionInfo('s', 's', 'i') },
	{ "bindtextdomain", FunctionInfo('s', 's') },
	{ "textdomain", FunctionInfo('s') },
	{ "getopt", FunctionInfo('i', 'p', 's') },
	{ "getopt_long", FunctionInfo('i', 'p', 's', 'p', 'p') },
};

const ArchProcessor::Function *ArchProcessor::funcs_begin	= funcs;
const ArchProcessor::Function *ArchProcessor::funcs_end		= funcs + (sizeof(funcs) / sizeof(funcs[0]));

//------------------------------------------------------------------------------
// Name: ArchProcessor()
// Desc: 
//------------------------------------------------------------------------------
ArchProcessor::ArchProcessor() : m_SplitEFlags(0) {
}

//------------------------------------------------------------------------------
// Name: setupRegisterItem(QCatagoryList *catagoryList, QTreeWidgetItem *parent, Register reg)
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::setupRegisterItem(QCatagoryList *catagoryList, QTreeWidgetItem *parent, Register reg) {
	
	Q_CHECK_PTR(catagoryList);
	
	QTreeWidgetItem *const p = catagoryList->addItem(parent, QString());
	
	Q_CHECK_PTR(p);

	p->setData(0, 1, reg);
	m_RegisterViewItems.push_back(p);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::setupRegisterView(QCatagoryList *catagoryList) {

	Q_CHECK_PTR(catagoryList);

	// setup the register view
	QTreeWidgetItem *const gpr = catagoryList->addCatagory("General Purpose");
	
	Q_CHECK_PTR(gpr);
	
	setupRegisterItem(catagoryList, gpr, REG_EAX);
	setupRegisterItem(catagoryList, gpr, REG_EBX);
	setupRegisterItem(catagoryList, gpr, REG_ECX);
	setupRegisterItem(catagoryList, gpr, REG_EDX);
	setupRegisterItem(catagoryList, gpr, REG_EBP);
	setupRegisterItem(catagoryList, gpr, REG_ESP);
	setupRegisterItem(catagoryList, gpr, REG_ESI);
	setupRegisterItem(catagoryList, gpr, REG_EDI);
	setupRegisterItem(catagoryList, gpr, REG_EIP);
	setupRegisterItem(catagoryList, gpr, REG_EFLAGS);
	
	// split eflags view
	m_SplitEFlags = new QTreeWidgetItem(m_RegisterViewItems[9]);	
	m_SplitEFlags->setText(0, "c p a z s d o");
	
	QTreeWidgetItem *const segs = catagoryList->addCatagory("Segments");
	
	Q_CHECK_PTR(segs);
	
	setupRegisterItem(catagoryList, segs, REG_CS);
	setupRegisterItem(catagoryList, segs, REG_DS);
	setupRegisterItem(catagoryList, segs, REG_ES);
	setupRegisterItem(catagoryList, segs, REG_FS);
	setupRegisterItem(catagoryList, segs, REG_GS);
	setupRegisterItem(catagoryList, segs, REG_SS);

	QTreeWidgetItem *const fpu = catagoryList->addCatagory("FPU");
	
	Q_CHECK_PTR(fpu);
	
	setupRegisterItem(catagoryList, fpu, REG_ST0);
	setupRegisterItem(catagoryList, fpu, REG_ST1);
	setupRegisterItem(catagoryList, fpu, REG_ST2);
	setupRegisterItem(catagoryList, fpu, REG_ST3);
	setupRegisterItem(catagoryList, fpu, REG_ST4);
	setupRegisterItem(catagoryList, fpu, REG_ST5);
	setupRegisterItem(catagoryList, fpu, REG_ST6);
	setupRegisterItem(catagoryList, fpu, REG_ST7);

	updateRegisterView(QString());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::reg_t *ArchProcessor::valueFromItem(const QTreeWidgetItem &item, unsigned int mode) {

	const Register reg = static_cast<Register>(item.data(0, 1).toUInt());
	edb::reg_t *ret = 0;
	
	State &state = edb::v1::currentState();
	
	switch(reg) {
	case REG_EAX:		if(mode & State::REG_GPR) ret = &state.eax; break;
	case REG_EBX:		if(mode & State::REG_GPR) ret = &state.ebx; break;
	case REG_ECX:		if(mode & State::REG_GPR) ret = &state.ecx; break;
	case REG_EDX:		if(mode & State::REG_GPR) ret = &state.edx; break;
	case REG_EBP:		if(mode & State::REG_GPR) ret = &state.ebp; break;
	case REG_ESP:		if(mode & State::REG_GPR) ret = &state.esp; break;
	case REG_ESI:		if(mode & State::REG_GPR) ret = &state.esi; break;
	case REG_EDI:		if(mode & State::REG_GPR) ret = &state.edi; break;
	case REG_EFLAGS:	if(mode & State::REG_COND) ret = &state.eflags; break;
	case REG_EIP:		if(mode & State::REG_IP) ret = &state.eip; break;
	case REG_CS:		if(mode & State::REG_SEG) ret = &state._cs; break;
	case REG_DS:		if(mode & State::REG_SEG) ret = &state._ds; break;
	case REG_ES:		if(mode & State::REG_SEG) ret = &state._es; break;
	case REG_FS:		if(mode & State::REG_SEG) ret = &state._fs; break;
	case REG_GS:		if(mode & State::REG_SEG) ret = &state._gs; break;
	case REG_SS:		if(mode & State::REG_SEG) ret = &state._ss; break;
	/*
	case REG_ST0:		if(mode & State::REG_FPU) ret = &state.st[0]; break;
	case REG_ST1:		if(mode & State::REG_FPU) ret = &state.st[1]; break;
	case REG_ST2:		if(mode & State::REG_FPU) ret = &state.st[2]; break;
	case REG_ST3:		if(mode & State::REG_FPU) ret = &state.st[3]; break;
	case REG_ST4:		if(mode & State::REG_FPU) ret = &state.st[4]; break;
	case REG_ST5:		if(mode & State::REG_FPU) ret = &state.st[5]; break;
	case REG_ST6		if(mode & State::REG_FPU) ret = &state.st[6]; break;
	case REG_ST7:		if(mode & State::REG_FPU) ret = &state.st[7]; break;
	*/
	default:
		break;
	}

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QTreeWidgetItem *ArchProcessor::getRegisterItem(unsigned int index) {
	Q_ASSERT(index < static_cast<unsigned int>(m_RegisterViewItems.size()));
	return m_RegisterViewItems[index];
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::doUpdateRegister(QTreeWidgetItem *item, const QString &name, edb::reg_t value) const {
	
	Q_CHECK_PTR(item);

	QString regString;
	QString text;
	int stringLen;
	
	// TODO: uses elided text for ascii string?
	if(edb::v1::getASCIIStringAtAddress(value, regString, edb::v1::config().minAsciiStringLength, 128, stringLen)) {
		text.sprintf("%s: " EDB_FMT_PTR " ASCII \"%s\"", qPrintable(name), value, qPrintable(regString));
	} else {
		text.sprintf("%s: " EDB_FMT_PTR, qPrintable(name), value);
	}
	
	item->setText(0, text);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::reset() {
	m_LastState.clear();
	m_SplitEFlags->setText(0, "c p a z s d o");
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::updateRegisterView(const QString &defaultRegionName) {

	const State &state = edb::v1::currentState();

	doUpdateRegister(getRegisterItem(0), "EAX", state.eax);
	doUpdateRegister(getRegisterItem(1), "EBX", state.ebx);
	doUpdateRegister(getRegisterItem(2), "ECX", state.ecx);
	doUpdateRegister(getRegisterItem(3), "EDX", state.edx);
	doUpdateRegister(getRegisterItem(4), "EBP", state.ebp);
	doUpdateRegister(getRegisterItem(5), "ESP", state.esp);
	doUpdateRegister(getRegisterItem(6), "ESI", state.esi);
	doUpdateRegister(getRegisterItem(7), "EDI", state.edi);

	const QString symname = edb::v1::findFunctionSymbol(state.eip, defaultRegionName);
	QString temp;
	getRegisterItem(8)->setText(0, temp.sprintf("EIP: " EDB_FMT_PTR " <%s>", state.eip, qPrintable(symname)));
	getRegisterItem(9)->setText(0, temp.sprintf("EFLAGS: " EDB_FMT_PTR, state.eflags));

	getRegisterItem(10)->setText(0, temp.sprintf("CS: %04x", static_cast<quint16>(state.cs) & 0xffff));
	getRegisterItem(11)->setText(0, temp.sprintf("DS: %04x", static_cast<quint16>(state.ds) & 0xffff));
	getRegisterItem(12)->setText(0, temp.sprintf("ES: %04x", static_cast<quint16>(state.es) & 0xffff));
	getRegisterItem(13)->setText(0, temp.sprintf("FS: %04x", static_cast<quint16>(state.fs) & 0xffff));
	getRegisterItem(14)->setText(0, temp.sprintf("GS: %04x", static_cast<quint16>(state.gs) & 0xffff));
	getRegisterItem(15)->setText(0, temp.sprintf("SS: %04x", static_cast<quint16>(state.ss) & 0xffff));

	// FPU
	for(int i = 0; i < 8; ++i) {
		getRegisterItem(16 + i)->setText(0, temp.sprintf("ST%d: %.16lg", i, state.st[i].toDouble()));
		getRegisterItem(16 + i)->setForeground(0, (state.st[i] != m_LastState.st[i]) ? Qt::red : Qt::black);
	}

	const bool eflagsChanged = state.eflags != m_LastState.eflags;
	if(eflagsChanged) {
		QString eflagString;
		
		eflagString += ((state.eflags & 0x001) ? "C " : "c ");
		eflagString += ((state.eflags & 0x004) ? "P " : "p ");
		eflagString += ((state.eflags & 0x010) ? "A " : "a ");
		eflagString += ((state.eflags & 0x040) ? "Z " : "z ");
		eflagString += ((state.eflags & 0x080) ? "S " : "s ");
		eflagString += ((state.eflags & 0x400) ? "D " : "d ");
		eflagString += ((state.eflags & 0x800) ? "O" : "o");

		m_SplitEFlags->setText(0, eflagString);
	}

	// highlight any changed registers
	getRegisterItem(0)->setForeground(0, (state.eax != m_LastState.eax) ? Qt::red : Qt::black);
	getRegisterItem(1)->setForeground(0, (state.ebx != m_LastState.ebx) ? Qt::red : Qt::black);
	getRegisterItem(2)->setForeground(0, (state.ecx != m_LastState.ecx) ? Qt::red : Qt::black);
	getRegisterItem(3)->setForeground(0, (state.edx != m_LastState.edx) ? Qt::red : Qt::black);
	getRegisterItem(4)->setForeground(0, (state.ebp != m_LastState.ebp) ? Qt::red : Qt::black);
	getRegisterItem(5)->setForeground(0, (state.esp != m_LastState.esp) ? Qt::red : Qt::black);
	getRegisterItem(6)->setForeground(0, (state.esi != m_LastState.esi) ? Qt::red : Qt::black);
	getRegisterItem(7)->setForeground(0, (state.edi != m_LastState.edi) ? Qt::red : Qt::black);
	getRegisterItem(8)->setForeground(0, (state.eip != m_LastState.eip) ? Qt::red : Qt::black);
	getRegisterItem(9)->setForeground(0, eflagsChanged ? Qt::red : Qt::black);

	state.toBackup(m_LastState);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::doFunctionParamterResolution(const QString &symname, int offset, QStringList &ret) const {

	// we will always be removing the last 2 chars "+0" from the string as well
	// as chopping the region prefix we like to prepend to symbols
	QString funcName;
	const int indexColon = symname.indexOf("::");

	if(indexColon != -1) {
		funcName = symname.left(symname.length() - 2).mid(indexColon + 2);
	}

	// safe not to check for -1, it means "rest of string" for the mid function
	funcName = funcName.mid(0, funcName.indexOf("@"));
	
	const Function *const it = qFind(funcs_begin, funcs_end, funcName);
	if(it != funcs_end) {
	
		const FunctionInfo *const info = &it->f;

		QString functionCall = funcName + "(";
		
		const QVector<QChar> params(info->params());
		int i = 0;
		foreach(QChar ch, params) {
		
			edb::reg_t arg;
			edb::v1::debuggerBase->readBytes(edb::v1::currentState().esp + i * sizeof(edb::reg_t) + offset, &arg, sizeof(arg));
			
			QString paramText;
			
			switch(ch.toAscii()) {
			case 's':
			case 'f':
				do {
					QString stringParam;
					int stringLen;
					
					if(edb::v1::getASCIIStringAtAddress(arg, stringParam, edb::v1::config().minAsciiStringLength, INT_MAX, stringLen)) {
						paramText.sprintf("<0x" EDB_FMT_PTR "> \"%s\"", arg, qPrintable(stringParam));
					} else {
						char character;
						edb::v1::debuggerBase->readBytes(arg, &character, sizeof(character));
						if(character == '\0') {
							paramText.sprintf("<0x" EDB_FMT_PTR "> \"\"", arg);
						} else {						
							paramText.sprintf("0x" EDB_FMT_PTR, arg);
						}
					}
				} while(0);
				break;
					
			case 'p':
				paramText.sprintf("0x" EDB_FMT_PTR, arg);
				break;
			case 'c':
				if(safe_ctype<std::isprint>(arg) || arg == ' ') {
					paramText.sprintf("'%c'", static_cast<char>(arg));
				} else {
					paramText.sprintf("'\\x%02x'", static_cast<quint16>(arg));
				}
				break;
			case 'u':
				paramText.sprintf("%u", static_cast<unsigned int>(arg));
				break;
			case 'i':
				paramText.sprintf("%d", static_cast<int>(arg));
				break;
			}
			
			functionCall += paramText;
			
			if(i + 1 < params.size()) {
				functionCall += ", ";
			}
			++i;
		}
		
		functionCall += ")";
		ret << functionCall;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
bool ArchProcessor::jccTaken(quint8 instructionByte) const {
	
	const edb::reg_t efl = edb::v1::currentState().eflags;
	const bool cf = (efl & 0x0001) != 0;
	const bool pf = (efl & 0x0004) != 0;
	const bool zf = (efl & 0x0040) != 0;
	const bool sf = (efl & 0x0080) != 0;
	const bool of = (efl & 0x0800) != 0;
	
	bool taken = false;
	
	switch(instructionByte & 0x0e) {
	case 0x00:
		taken = of;
		break;
	case 0x02:
		taken = cf;
		break;
	case 0x04:
		taken = zf;
		break;
	case 0x06:
		taken = cf || zf;
		break;
	case 0x08:
		taken = sf;
		break;
	case 0x0a:
		taken = pf;
		break;
	case 0x0c:
		taken = sf != of;
		break;
	case 0x0e:
		taken = zf || sf != of;
		break;
	}
	
	if(instructionByte & 0x01) {
		taken = !taken;
	}
	
	return taken;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeCMov(const Instruction &insn, QStringList &ret) const {
	
	const quint8 *const insn_buf = insn.buffer();

	const bool taken = jccTaken(insn_buf[1 + insn.prefixSize()]);
	
	if(taken) {
		ret << "move performed";
	} else {
		ret << "move NOT performed";
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeJump(const Instruction &insn, QStringList &ret) const {

	bool taken = false;
	
	const quint8 *const insn_buf = insn.buffer();
	
	if(insn.type() == Instruction::OP_JCC) {
		
		if(insn.size() - insn.prefixSize() == 2) {
			taken = jccTaken(insn_buf[0 + insn.prefixSize()]);
		} else {
			taken = jccTaken(insn_buf[1 + insn.prefixSize()]);
		}
		
	} else if(insn_buf[0] == 0xe3) {
		if(insn.prefix() & Instruction::PREFIX_ADDRESS) {
			taken = (edb::v1::currentState().ecx & 0xffff) == 0;
		} else {
			taken = edb::v1::currentState().ecx == 0;
		}
	}
	
	if(taken) {
		ret << "jump taken";
	} else {
		ret << "jump NOT taken";
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeReturn(const Instruction &insn, QStringList &ret) const {

	Q_UNUSED(insn);

	edb::address_t returnAddress;
	edb::v1::debuggerBase->readBytes(edb::v1::currentState().esp, &returnAddress, sizeof(returnAddress));

	const QString symname = edb::v1::findFunctionSymbol(returnAddress, QString());
	QString text;
	if(!symname.isEmpty()) {
		text.sprintf("Return to " EDB_FMT_PTR " <%s>", returnAddress, qPrintable(symname));
	} else {
		text.sprintf("Return to " EDB_FMT_PTR, returnAddress);
	}
	
	ret << text;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t ArchProcessor::getEffectiveAddress(const Operand &op) const {
	return getEffectiveAddress(op, edb::v1::currentState());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t ArchProcessor::getEffectiveAddress(const Operand &op, const State &state) const {
	edb::address_t ret = 0;
	
	// TODO: get registers by index, not string! too slow
	
	if(op.valid()) {
	
		Instruction *const owner = op.owner();
		
		Q_CHECK_PTR(owner);

		switch(op.generalType()) {

		case Operand::TYPE_REGISTER:
			ret = state.value(op.registerName(op.reg).c_str());
			break;
		case Operand::TYPE_EXPRESSION:
			do {
				const edb::reg_t base	= state.value(op.registerName(op.expression.base).c_str());
				const edb::reg_t index	= state.value(op.registerName(op.expression.index).c_str());
				ret						= base + index * op.expression.scale + op.displacement();
			} while(0);
			break;
		case Operand::TYPE_ABSOLUTE:
			// TODO: find a way to take into account the segment base/limit
			ret = op.absolute.offset;
			break;
		case Operand::TYPE_IMMEDIATE:
			break;
		case Operand::TYPE_REL:
			if(op.completeType() == Operand::TYPE_REL8) {
				ret = owner->rva32() + owner->size() + op.sbyte;
			} else if(op.completeType() == Operand::TYPE_REL32) {
				ret = owner->rva32() + owner->size() + op.sdword;
			} else if(op.completeType() == Operand::TYPE_REL16) {
				// NOTE: intel truncates EIP to 16-bit
				ret = (owner->rva32() + owner->size() + op.sword) & 0xffff;
			}
			break;	
		default:
			break;
		}
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeCall(const Instruction &insn, QStringList &ret) const {
	
	const Operand &operand = insn.operand(0);
	
	if(operand.valid()) {

		const edb::address_t effective_address	= getEffectiveAddress(operand);
		const QString tempOperand				= operand.format().c_str();
		QString temp;

		switch(operand.generalType()) {
		case Operand::TYPE_REL:
		case Operand::TYPE_REGISTER:
			do {
				int offset;
				const QString symname = edb::v1::findFunctionSymbol(effective_address, QString(), &offset);
				if(!symname.isEmpty()) {
					ret << temp.sprintf("%s = " EDB_FMT_PTR " <%s>", qPrintable(tempOperand), effective_address, qPrintable(symname));

					if(offset == 0) {
						if(insn.type() == Instruction::OP_CALL) {
							doFunctionParamterResolution(symname, 0, ret);
						} else {
							doFunctionParamterResolution(symname, 4, ret);
						}
					}

				} else {
					ret << temp.sprintf("%s = " EDB_FMT_PTR, qPrintable(tempOperand), effective_address);
				}
			} while(0);
			break;

		case Operand::TYPE_EXPRESSION:
		default:
			do {
				edb::address_t target;
				const bool ok = edb::v1::debuggerBase->readBytes(effective_address, &target, sizeof(target));

				if(ok) {
					int offset;
					const QString symname = edb::v1::findFunctionSymbol(target, QString(), &offset);
					if(!symname.isEmpty()) {
						ret << temp.sprintf("%s = [" EDB_FMT_PTR "] = " EDB_FMT_PTR " <%s>", qPrintable(tempOperand), effective_address, target, qPrintable(symname));

						if(offset == 0) {
							if(insn.type() == Instruction::OP_CALL) {
								doFunctionParamterResolution(symname, 0, ret);
							} else {
								doFunctionParamterResolution(symname, 4, ret);
							}
						}

					} else {
						ret << temp.sprintf("%s = [" EDB_FMT_PTR "] = " EDB_FMT_PTR, qPrintable(tempOperand), effective_address, target);
					}
				} else {
					// could not read from the address
					ret << temp.sprintf("%s = [" EDB_FMT_PTR "] = ?", qPrintable(tempOperand), effective_address);
				}
			} while(0);
			break;
		}
	}
}

//------------------------------------------------------------------------------
// Name: analyzeOperands(const Instruction &insn, QStringList &ret)
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeOperands(const Instruction &insn, QStringList &ret) const {

	Q_UNUSED(insn);
	
	for(int j = 0; j < Instruction::MAX_OPERANDS; ++j) {
	
		const Operand &operand = insn.operand(j);

		if(operand.valid()) {

			const QString tempOperand = operand.format().c_str();
			QString text;
			
			switch(operand.generalType()) {
			case Operand::TYPE_REL:
			case Operand::TYPE_REGISTER:
				do {
					const edb::address_t effective_address = getEffectiveAddress(operand);
					ret << text.sprintf("%s = " EDB_FMT_PTR, qPrintable(tempOperand), effective_address);
				} while(0);
				break;
			case Operand::TYPE_EXPRESSION:
				do {
					const edb::address_t effective_address = getEffectiveAddress(operand);
					edb::address_t target;

					const bool ok = edb::v1::debuggerBase->readBytes(effective_address, &target, sizeof(target));

					
					if(ok) {
						
						switch(operand.completeType()) {
						case Operand::TYPE_EXPRESSION8:
							ret << text.sprintf("%s = [" EDB_FMT_PTR "] = %02x", qPrintable(tempOperand), effective_address, static_cast<quint8>(target) & 0xff);
							break;
						case Operand::TYPE_EXPRESSION16:
							ret << text.sprintf("%s = [" EDB_FMT_PTR "] = %04x", qPrintable(tempOperand), effective_address, static_cast<quint16>(target) & 0xffff);
							break;
						case Operand::TYPE_EXPRESSION32:
						default:
							ret << text.sprintf("%s = [" EDB_FMT_PTR "] = %08x", qPrintable(tempOperand), effective_address, static_cast<quint32>(target));
							break;
						}
					} else {
						ret << text.sprintf("%s = [" EDB_FMT_PTR "] = ?", qPrintable(tempOperand), effective_address);
					}
				} while(0);
				break;
			default:
				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: analyzeJumpTargets(const Instruction &insn, QStringList &ret) const
// Desc: 
//------------------------------------------------------------------------------
void ArchProcessor::analyzeJumpTargets(const Instruction &insn, QStringList &ret) const {
	const edb::address_t address		= insn.rva32();
	const edb::address_t start_address	= address - 128;
	const edb::address_t end_address	= address + 128;
	
	quint8 buffer[Instruction::MAX_SIZE];
	
	for(edb::address_t addr = start_address; addr < end_address; ++addr) {
		int sz = sizeof(buffer);
		const bool ok = edb::v1::getInstructionBytes(addr, buffer, sz);
		if(ok) {
			Instruction insn(buffer, sz, addr, std::nothrow);
			if(insn.valid() && insn.type() == Instruction::OP_JCC || insn.type() == Instruction::OP_JMP) {
				const Operand &operand = insn.operand(0);
				
				if(operand.generalType() == Operand::TYPE_REL) {
					const edb::address_t target = operand.relativeTarget();
					
					if(target == address) {
						ret << QString().sprintf("possible jump from " EDB_FMT_PTR, addr);
					}
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: updateInstructionInfo(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
QStringList ArchProcessor::updateInstructionInfo(edb::address_t address) {

	QStringList ret;
	
	Q_CHECK_PTR(edb::v1::debuggerBase);
	
	quint8 buffer[Instruction::MAX_SIZE];

	const bool ok = edb::v1::debuggerBase->readBytes(address, buffer, sizeof(buffer));
	if(ok) {
		Instruction insn(buffer, sizeof(buffer), address, std::nothrow);
		if(insn.valid()) {
			// figure out the instruction type and display some information about it
			switch(insn.type()) {
			case Instruction::OP_CMOVCC:
				analyzeCMov(insn, ret);
				break;
			case Instruction::OP_RET:
				analyzeReturn(insn, ret);
				break;
			case Instruction::OP_JCC:
				analyzeJump(insn, ret);
				// FALL THROUGH!
			case Instruction::OP_JMP:
			case Instruction::OP_CALL:
				analyzeCall(insn, ret);
				break;
			case Instruction::OP_SYSENTER:
				break;
			default:
				analyzeOperands(insn, ret);
				break;
			}
			
			analyzeJumpTargets(insn, ret);
			
		}
	}
	
	// eliminate duplicates
	ret = QStringList::fromSet(ret.toSet());
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: canStepOver(const Instruction &insn) const
// Desc: 
//------------------------------------------------------------------------------
bool ArchProcessor::canStepOver(const Instruction &insn) const {
	return insn.valid() && (insn.type() == Instruction::OP_CALL || (insn.prefix() & (Instruction::PREFIX_REPNE | Instruction::PREFIX_REP)));
}

//------------------------------------------------------------------------------
// Name: isFilling(const Instruction &insn, const QByteArray &bytes) const
// Desc: 
//------------------------------------------------------------------------------
bool ArchProcessor::isFilling(const Instruction &insn, const QByteArray &bytes) const {
	bool ret = false;
	
	// fetch the operands
	const Operand operands[Instruction::MAX_OPERANDS] = {
		insn.operand(0),
		insn.operand(1),
		insn.operand(2)
	};
	
	switch(insn.type()) {
	case Instruction::OP_NOP:
	case Instruction::OP_INT3:
		ret = true;
		break;
		
	case Instruction::OP_LEA:
		if(operands[0].valid() && operands[1].valid()) {
			if(operands[0].generalType() == Operand::TYPE_REGISTER && operands[1].generalType() == Operand::TYPE_EXPRESSION) {

				Operand::Register reg1;
				Operand::Register reg2;

				reg1 = operands[0].reg;

				if(operands[1].expression.scale == 1) {
					if(operands[1].expression.u_disp32 == 0) {

						if(operands[1].expression.base == Operand::REG_NONE) {
							reg2 = operands[1].expression.index;
							ret = reg1 == reg2;
						} else if(operands[1].expression.index == Operand::REG_NONE) {
							reg2 = operands[1].expression.base;
							ret = reg1 == reg2;
						}
					}
				}
			}
		}
		break;
		
	case Instruction::OP_MOV:
		if(operands[0].valid() && operands[1].valid()) {
			if(operands[0].generalType() == Operand::TYPE_REGISTER && operands[1].generalType() == Operand::TYPE_REGISTER) {
				ret = operands[0].reg == operands[1].reg;
			}
		}
		break;
		
	default:
		break;
	}
	
	if(!ret) {
		if(edb::v1::config().zerosAreFilling) {			
			ret = (bytes == QByteArray::fromRawData("\x00\x00", 2));
		}
	}
	
	return ret;
}
